from enum import Enum
from typing import Callable, Coroutine, Any, Type, Set
from typing import (
    Dict,
    List,
    Optional,
    Sequence,
    Union,
)

from fastapi import APIRouter
from fastapi import params
from fastapi.datastructures import Default, DefaultPlaceholder
from fastapi.exceptions import FastAPIError
from fastapi.routing import APIRoute, APIWebSocketRoute
from fastapi.types import DecoratedCallable, IncEx
from fastapi.utils import (
    generate_unique_id, get_value_or_default,
)
from starlette import routing
from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.routing import (
    BaseRoute,
)
from typing_extensions import Annotated, Doc

from cachepot.constants import CachePolicy
from cachepot.utils import get_request_handler


class CachedAPIRoute(APIRoute):
    def __init__(self, *args, cache_policy, **kwargs):
        self.cache_policy = cache_policy
        super().__init__(*args, **kwargs)

    def get_route_handler(self) -> Callable[[Request], Coroutine[Any, Any, Response]]:
        return get_request_handler(
            dependant=self.dependant,
            body_field=self.body_field,
            status_code=self.status_code,
            response_class=self.response_class,
            response_field=self.secure_cloned_response_field,
            response_model_include=self.response_model_include,
            response_model_exclude=self.response_model_exclude,
            response_model_by_alias=self.response_model_by_alias,
            response_model_exclude_unset=self.response_model_exclude_unset,
            response_model_exclude_defaults=self.response_model_exclude_defaults,
            response_model_exclude_none=self.response_model_exclude_none,
            dependency_overrides_provider=self.dependency_overrides_provider,
            cache_policy=self.cache_policy
        )


class CachedAPIRouter(APIRouter):
    def __init__(
            self,
            *args,
            route_class: Annotated[
                Type[APIRoute],
                Doc(
                    """
                    Custom route (*path operation*) class to be used by this router.

                    Read more about it in the
                    [FastAPI docs for Custom Request and APIRoute class](https://fastapi.tiangolo.com/how-to/
                    custom-request-and-route/#custom-apiroute-class-in-a-router).
                    """
                ),
            ] = CachedAPIRoute,
            **kwargs,
    ):
        super().__init__(*args, route_class=route_class, **kwargs)

    def include_router(
            self,
            router: Annotated['APIRouter', Doc('The `APIRouter` to include.')],
            *,
            prefix: Annotated[str, Doc('An optional path prefix for the router.')] = '',
            tags: Annotated[
                Optional[List[Union[str, Enum]]],
                Doc(
                    """
                    A list of tags to be applied to all the *path operations* in this
                    router.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                    Read more about it in the
                    [FastAPI docs for Path Operation Configuration](https://fastapi.tiangolo.com/tutorial/
                    path-operation-configuration/).
                    """
                ),
            ] = None,
            dependencies: Annotated[
                Optional[Sequence[params.Depends]],
                Doc(
                    """
                    A list of dependencies (using `Depends()`) to be applied to all the
                    *path operations* in this router.

                    Read more about it in the
                    [FastAPI docs for Bigger Applications - Multiple Files](https://fastapi.tiangolo.com/tutorial/
                    bigger-applications/#include-an-apirouter-with-a-custom-prefix-tags-responses-and-dependencies).
                    """
                ),
            ] = None,
            default_response_class: Annotated[
                Type[Response],
                Doc(
                    """
                    The default response class to be used.

                    Read more in the
                    [FastAPI docs for Custom Response - HTML, Stream, File, others](https://fastapi.tiangolo.com/
                    advanced/custom-response/#default-response-class).
                    """
                ),
            ] = Default(JSONResponse),
            responses: Annotated[
                Optional[Dict[Union[int, str], Dict[str, Any]]],
                Doc(
                    """
                    Additional responses to be shown in OpenAPI.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                    Read more about it in the
                    [FastAPI docs for Additional Responses in OpenAPI](https://fastapi.tiangolo.com/advanced/
                    additional-responses/).

                    And in the
                    [FastAPI docs for Bigger Applications](https://fastapi.tiangolo.com/tutorial/bigger-applications/
                    #include-an-apirouter-with-a-custom-prefix-tags-responses-and-dependencies).
                    """
                ),
            ] = None,
            callbacks: Annotated[
                Optional[List[BaseRoute]],
                Doc(
                    """
                    OpenAPI callbacks that should apply to all *path operations* in this
                    router.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                    Read more about it in the
                    [FastAPI docs for OpenAPI Callbacks](https://fastapi.tiangolo.com/advanced/openapi-callbacks/).
                    """
                ),
            ] = None,
            deprecated: Annotated[
                Optional[bool],
                Doc(
                    """
                    Mark all *path operations* in this router as deprecated.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                    Read more about it in the
                    [FastAPI docs for Path Operation Configuration](https://fastapi.tiangolo.com/tutorial/
                    path-operation-configuration/).
                    """
                ),
            ] = None,
            include_in_schema: Annotated[
                bool,
                Doc(
                    """
                    Include (or not) all the *path operations* in this router in the
                    generated OpenAPI schema.

                    This affects the generated OpenAPI (e.g. visible at `/docs`).
                    """
                ),
            ] = True,
            generate_unique_id_function: Annotated[
                Callable[[APIRoute], str],
                Doc(
                    """
                    Customize the function used to generate unique IDs for the *path
                    operations* shown in the generated OpenAPI.

                    This is particularly useful when automatically generating clients or
                    SDKs for your API.

                    Read more about it in the
                    [FastAPI docs about how to Generate Clients](https://fastapi.tiangolo.com/advanced/generate-clients
                    /#custom-generate-unique-id-function).
                    """
                ),
            ] = Default(generate_unique_id),
    ) -> None:
        """
        Include another `APIRouter` in the same current `APIRouter`.

        Read more about it in the
        [FastAPI docs for Bigger Applications](https://fastapi.tiangolo.com/tutorial/bigger-applications/).

        ## Example

        ```python
        from fastapi import APIRouter, FastAPI

        app = FastAPI()
        internal_router = APIRouter()
        users_router = APIRouter()

        @users_router.get('/users/')
        def read_users():
            return [{'name': 'Rick'}, {'name': 'Morty'}]

        internal_router.include_router(users_router)
        app.include_router(internal_router)
        ```
        """
        if prefix:
            assert prefix.startswith('/'), 'A path prefix must start with ' / ''
            assert not prefix.endswith(
                '/'
            ), 'A path prefix must not end with ' / ', as the routes will start with ' / ''
        else:
            for r in router.routes:
                path = getattr(r, 'path')  # noqa: B009
                name = getattr(r, 'name', 'unknown')
                if path is not None and not path:
                    raise FastAPIError(
                        f'Prefix and path cannot be both empty (path operation: {name})'
                    )
        if responses is None:
            responses = {}
        for route in router.routes:
            if isinstance(route, APIRoute):
                combined_responses = {**responses, **route.responses}
                use_response_class = get_value_or_default(
                    route.response_class,
                    router.default_response_class,
                    default_response_class,
                    self.default_response_class,
                )
                current_tags = []
                if tags:
                    current_tags.extend(tags)
                if route.tags:
                    current_tags.extend(route.tags)
                current_dependencies: List[params.Depends] = []
                if dependencies:
                    current_dependencies.extend(dependencies)
                if route.dependencies:
                    current_dependencies.extend(route.dependencies)
                current_callbacks = []
                if callbacks:
                    current_callbacks.extend(callbacks)
                if route.callbacks:
                    current_callbacks.extend(route.callbacks)
                current_generate_unique_id = get_value_or_default(
                    route.generate_unique_id_function,
                    router.generate_unique_id_function,
                    generate_unique_id_function,
                    self.generate_unique_id_function,
                )
                if isinstance(route, CachedAPIRoute):
                    self.add_api_route(
                        prefix + route.path,
                        route.endpoint,
                        response_model=route.response_model,
                        status_code=route.status_code,
                        tags=current_tags,
                        dependencies=current_dependencies,
                        summary=route.summary,
                        description=route.description,
                        response_description=route.response_description,
                        responses=combined_responses,
                        deprecated=route.deprecated or deprecated or self.deprecated,
                        methods=route.methods,
                        operation_id=route.operation_id,
                        response_model_include=route.response_model_include,
                        response_model_exclude=route.response_model_exclude,
                        response_model_by_alias=route.response_model_by_alias,
                        response_model_exclude_unset=route.response_model_exclude_unset,
                        response_model_exclude_defaults=route.response_model_exclude_defaults,
                        response_model_exclude_none=route.response_model_exclude_none,
                        include_in_schema=route.include_in_schema and self.include_in_schema and include_in_schema,
                        response_class=use_response_class,
                        name=route.name,
                        route_class_override=type(route),
                        callbacks=current_callbacks,
                        openapi_extra=route.openapi_extra,
                        generate_unique_id_function=current_generate_unique_id,
                        cache_policy=route.cache_policy,
                    )
                else:
                    self.add_api_route(
                        prefix + route.path,
                        route.endpoint,
                        response_model=route.response_model,
                        status_code=route.status_code,
                        tags=current_tags,
                        dependencies=current_dependencies,
                        summary=route.summary,
                        description=route.description,
                        response_description=route.response_description,
                        responses=combined_responses,
                        deprecated=route.deprecated or deprecated or self.deprecated,
                        methods=route.methods,
                        operation_id=route.operation_id,
                        response_model_include=route.response_model_include,
                        response_model_exclude=route.response_model_exclude,
                        response_model_by_alias=route.response_model_by_alias,
                        response_model_exclude_unset=route.response_model_exclude_unset,
                        response_model_exclude_defaults=route.response_model_exclude_defaults,
                        response_model_exclude_none=route.response_model_exclude_none,
                        include_in_schema=route.include_in_schema and self.include_in_schema and include_in_schema,
                        response_class=use_response_class,
                        name=route.name,
                        route_class_override=type(route),
                        callbacks=current_callbacks,
                        openapi_extra=route.openapi_extra,
                        generate_unique_id_function=current_generate_unique_id,
                    )
            elif isinstance(route, routing.Route):
                methods = list(route.methods or [])
                self.add_route(
                    prefix + route.path,
                    route.endpoint,
                    methods=methods,
                    include_in_schema=route.include_in_schema,
                    name=route.name,
                )
            elif isinstance(route, APIWebSocketRoute):
                current_dependencies = []
                if dependencies:
                    current_dependencies.extend(dependencies)
                if route.dependencies:
                    current_dependencies.extend(route.dependencies)
                self.add_api_websocket_route(
                    prefix + route.path,
                    route.endpoint,
                    dependencies=current_dependencies,
                    name=route.name,
                )
            elif isinstance(route, routing.WebSocketRoute):
                self.add_websocket_route(
                    prefix + route.path, route.endpoint, name=route.name
                )
        for handler in router.on_startup:
            self.add_event_handler('startup', handler)
        for handler in router.on_shutdown:
            self.add_event_handler('shutdown', handler)

    def add_api_route(
            self,
            path: str,
            endpoint: Callable[..., Any],
            *,
            response_model: Any = Default(None),
            status_code: Optional[int] = None,
            tags: Optional[List[Union[str, Enum]]] = None,
            dependencies: Optional[Sequence[params.Depends]] = None,
            summary: Optional[str] = None,
            description: Optional[str] = None,
            response_description: str = 'Successful Response',
            responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
            deprecated: Optional[bool] = None,
            methods: Optional[Union[Set[str], List[str]]] = None,
            operation_id: Optional[str] = None,
            response_model_include: Optional[IncEx] = None,
            response_model_exclude: Optional[IncEx] = None,
            response_model_by_alias: bool = True,
            response_model_exclude_unset: bool = False,
            response_model_exclude_defaults: bool = False,
            response_model_exclude_none: bool = False,
            include_in_schema: bool = True,
            response_class: Union[Type[Response], DefaultPlaceholder] = Default(
                JSONResponse
            ),
            name: Optional[str] = None,
            route_class_override: Optional[Type[CachedAPIRoute]] = None,
            callbacks: Optional[List[BaseRoute]] = None,
            openapi_extra: Optional[Dict[str, Any]] = None,
            generate_unique_id_function: Union[
                Callable[[APIRoute], str], DefaultPlaceholder
            ] = Default(generate_unique_id),
            cache_policy: Optional[CachePolicy] = None,
    ) -> None:
        route_class = route_class_override or self.route_class
        responses = responses or {}
        combined_responses = {**self.responses, **responses}
        current_response_class = get_value_or_default(
            response_class, self.default_response_class
        )
        current_tags = self.tags.copy()
        if tags:
            current_tags.extend(tags)
        current_dependencies = self.dependencies.copy()
        if dependencies:
            current_dependencies.extend(dependencies)
        current_callbacks = self.callbacks.copy()
        if callbacks:
            current_callbacks.extend(callbacks)
        current_generate_unique_id = get_value_or_default(
            generate_unique_id_function, self.generate_unique_id_function
        )
        params = {
            'path': self.prefix + path,
            'endpoint': endpoint,
            'response_model': response_model,
            'status_code': status_code,
            'tags': current_tags,
            'dependencies': current_dependencies,
            'summary': summary,
            'description': description,
            'response_description': response_description,
            'responses': combined_responses,
            'deprecated': deprecated or self.deprecated,
            'methods': methods,
            'operation_id': operation_id,
            'response_model_include': response_model_include,
            'response_model_exclude': response_model_exclude,
            'response_model_by_alias': response_model_by_alias,
            'response_model_exclude_unset': response_model_exclude_unset,
            'response_model_exclude_defaults': response_model_exclude_defaults,
            'response_model_exclude_none': response_model_exclude_none,
            'include_in_schema': include_in_schema and self.include_in_schema,
            'response_class': current_response_class,
            'name': name,
            'dependency_overrides_provider': self.dependency_overrides_provider,
            'callbacks': current_callbacks,
            'openapi_extra': openapi_extra,
            'generate_unique_id_function': current_generate_unique_id,
        }
        if route_class == CachedAPIRoute:
            params['cache_policy'] = cache_policy
        route = route_class(**params)
        self.routes.append(route)

    def api_route(
            self,
            path: str,
            *,
            response_model: Any = Default(None),
            status_code: Optional[int] = None,
            tags: Optional[List[Union[str, Enum]]] = None,
            dependencies: Optional[Sequence[params.Depends]] = None,
            summary: Optional[str] = None,
            description: Optional[str] = None,
            response_description: str = 'Successful Response',
            responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
            deprecated: Optional[bool] = None,
            methods: Optional[List[str]] = None,
            operation_id: Optional[str] = None,
            response_model_include: Optional[IncEx] = None,
            response_model_exclude: Optional[IncEx] = None,
            response_model_by_alias: bool = True,
            response_model_exclude_unset: bool = False,
            response_model_exclude_defaults: bool = False,
            response_model_exclude_none: bool = False,
            include_in_schema: bool = True,
            response_class: Type[Response] = Default(JSONResponse),
            name: Optional[str] = None,
            callbacks: Optional[List[BaseRoute]] = None,
            openapi_extra: Optional[Dict[str, Any]] = None,
            generate_unique_id_function: Callable[[APIRoute], str] = Default(
                generate_unique_id
            ),
            cache_policy: Optional[CachePolicy] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self.add_api_route(
                path,
                func,
                response_model=response_model,
                status_code=status_code,
                tags=tags,
                dependencies=dependencies,
                summary=summary,
                description=description,
                response_description=response_description,
                responses=responses,
                deprecated=deprecated,
                methods=methods,
                operation_id=operation_id,
                response_model_include=response_model_include,
                response_model_exclude=response_model_exclude,
                response_model_by_alias=response_model_by_alias,
                response_model_exclude_unset=response_model_exclude_unset,
                response_model_exclude_defaults=response_model_exclude_defaults,
                response_model_exclude_none=response_model_exclude_none,
                include_in_schema=include_in_schema,
                response_class=response_class,
                name=name,
                callbacks=callbacks,
                openapi_extra=openapi_extra,
                generate_unique_id_function=generate_unique_id_function,
                cache_policy=cache_policy,
            )
            return func

        return decorator

    def get(
            self,
            path: Annotated[
                str,
                Doc(
                    """
                    The URL path to be used for this *path operation*.

                    For example, in `http://example.com/items`, the path is `/items`.
                    """
                ),
            ],
            *,
            response_model: Annotated[
                Any,
                Doc(
                    """
                    The type to use for the response.

                    It could be any valid Pydantic *field* type. So, it doesn't have to
                    be a Pydantic model, it could be other things, like a `list`, `dict`,
                    etc.

                    It will be used for:

                    * Documentation: the generated OpenAPI (and the UI at `/docs`) will
                        show it as the response (JSON Schema).
                    * Serialization: you could return an arbitrary object and the
                        `response_model` would be used to serialize that object into the
                        corresponding JSON.
                    * Filtering: the JSON sent to the client will only contain the data
                        (fields) defined in the `response_model`. If you returned an object
                        that contains an attribute `password` but the `response_model` does
                        not include that field, the JSON sent to the client would not have
                        that `password`.
                    * Validation: whatever you return will be serialized with the
                        `response_model`, converting any data as necessary to generate the
                        corresponding JSON. But if the data in the object returned is not
                        valid, that would mean a violation of the contract with the client,
                        so it's an error from the API developer. So, FastAPI will raise an
                        error and return a 500 error code (Internal Server Error).

                    Read more about it in the
                    [FastAPI docs for Response Model](https://fastapi.tiangolo.com/tutorial/response-model/).
                    """
                ),
            ] = Default(None),
            status_code: Annotated[
                Optional[int],
                Doc(
                    """
                    The default status code to be used for the response.

                    You could override the status code by returning a response directly.

                    Read more about it in the
                    [FastAPI docs for Response Status Code](https://fastapi.tiangolo.com/tutorial/
                    response-status-code/).
                    """
                ),
            ] = None,
            tags: Annotated[
                Optional[List[Union[str, Enum]]],
                Doc(
                    """
                    A list of tags to be applied to the *path operation*.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                    Read more about it in the
                    [FastAPI docs for Path Operation Configuration](https://fastapi.tiangolo.com/tutorial/
                    path-operation-configuration/#tags).
                    """
                ),
            ] = None,
            dependencies: Annotated[
                Optional[Sequence[params.Depends]],
                Doc(
                    """
                    A list of dependencies (using `Depends()`) to be applied to the
                    *path operation*.

                    Read more about it in the
                    [FastAPI docs for Dependencies in path operation decorators](https://fastapi.tiangolo.com/tutorial/
                    dependencies/dependencies-in-path-operation-decorators/).
                    """
                ),
            ] = None,
            summary: Annotated[
                Optional[str],
                Doc(
                    """
                    A summary for the *path operation*.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                    Read more about it in the
                    [FastAPI docs for Path Operation Configuration](https://fastapi.tiangolo.com/tutorial/
                    path-operation-configuration/).
                    """
                ),
            ] = None,
            description: Annotated[
                Optional[str],
                Doc(
                    """
                    A description for the *path operation*.

                    If not provided, it will be extracted automatically from the docstring
                    of the *path operation function*.

                    It can contain Markdown.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                    Read more about it in the
                    [FastAPI docs for Path Operation Configuration](https://fastapi.tiangolo.com/tutorial/
                    path-operation-configuration/).
                    """
                ),
            ] = None,
            response_description: Annotated[
                str,
                Doc(
                    """
                    The description for the default response.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).
                    """
                ),
            ] = 'Successful Response',
            responses: Annotated[
                Optional[Dict[Union[int, str], Dict[str, Any]]],
                Doc(
                    """
                    Additional responses that could be returned by this *path operation*.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).
                    """
                ),
            ] = None,
            deprecated: Annotated[
                Optional[bool],
                Doc(
                    """
                    Mark this *path operation* as deprecated.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).
                    """
                ),
            ] = None,
            operation_id: Annotated[
                Optional[str],
                Doc(
                    """
                    Custom operation ID to be used by this *path operation*.

                    By default, it is generated automatically.

                    If you provide a custom operation ID, you need to make sure it is
                    unique for the whole API.

                    You can customize the
                    operation ID generation with the parameter
                    `generate_unique_id_function` in the `FastAPI` class.

                    Read more about it in the
                    [FastAPI docs about how to Generate Clients](https://fastapi.tiangolo.com/advanced/generate-clients
                    /#custom-generate-unique-id-function).
                    """
                ),
            ] = None,
            response_model_include: Annotated[
                Optional[IncEx],
                Doc(
                    """
                    Configuration passed to Pydantic to include only certain fields in the
                    response data.

                    Read more about it in the
                    [FastAPI docs for Response Model - Return Type](https://fastapi.tiangolo.com/tutorial/response-model
                    /#response_model_include-and-response_model_exclude).
                    """
                ),
            ] = None,
            response_model_exclude: Annotated[
                Optional[IncEx],
                Doc(
                    """
                    Configuration passed to Pydantic to exclude certain fields in the
                    response data.

                    Read more about it in the
                    [FastAPI docs for Response Model - Return Type](https://fastapi.tiangolo.com/tutorial/response-model
                    /#response_model_include-and-response_model_exclude).
                    """
                ),
            ] = None,
            response_model_by_alias: Annotated[
                bool,
                Doc(
                    """
                    Configuration passed to Pydantic to define if the response model
                    should be serialized by alias when an alias is used.

                    Read more about it in the
                    [FastAPI docs for Response Model - Return Type](https://fastapi.tiangolo.com/tutorial/response-model
                    /#response_model_include-and-response_model_exclude).
                    """
                ),
            ] = True,
            response_model_exclude_unset: Annotated[
                bool,
                Doc(
                    """
                    Configuration passed to Pydantic to define if the response data
                    should have all the fields, including the ones that were not set and
                    have their default values. This is different from
                    `response_model_exclude_defaults` in that if the fields are set,
                    they will be included in the response, even if the value is the same
                    as the default.

                    When `True`, default values are omitted from the response.

                    Read more about it in the
                    [FastAPI docs for Response Model - Return Type](https://fastapi.tiangolo.com/tutorial/
                    response-model/#use-the-response_model_exclude_unset-parameter).
                    """
                ),
            ] = False,
            response_model_exclude_defaults: Annotated[
                bool,
                Doc(
                    """
                    Configuration passed to Pydantic to define if the response data
                    should have all the fields, including the ones that have the same value
                    as the default. This is different from `response_model_exclude_unset`
                    in that if the fields are set but contain the same default values,
                    they will be excluded from the response.

                    When `True`, default values are omitted from the response.

                    Read more about it in the
                    [FastAPI docs for Response Model - Return Type](https://fastapi.tiangolo.com/tutorial/response-model
                    /#use-the-response_model_exclude_unset-parameter).
                    """
                ),
            ] = False,
            response_model_exclude_none: Annotated[
                bool,
                Doc(
                    """
                    Configuration passed to Pydantic to define if the response data should
                    exclude fields set to `None`.

                    This is much simpler (less smart) than `response_model_exclude_unset`
                    and `response_model_exclude_defaults`. You probably want to use one of
                    those two instead of this one, as those allow returning `None` values
                    when it makes sense.

                    Read more about it in the
                    [FastAPI docs for Response Model - Return Type](https://fastapi.tiangolo.com/tutorial/
                    response-model/#response_model_exclude_none).
                    """
                ),
            ] = False,
            include_in_schema: Annotated[
                bool,
                Doc(
                    """
                    Include this *path operation* in the generated OpenAPI schema.

                    This affects the generated OpenAPI (e.g. visible at `/docs`).

                    Read more about it in the
                    [FastAPI docs for Query Parameters and String Validations](https://fastapi.tiangolo.com/tutorial/
                    query-params-str-validations/#exclude-from-openapi).
                    """
                ),
            ] = True,
            response_class: Annotated[
                Type[Response],
                Doc(
                    """
                    Response class to be used for this *path operation*.

                    This will not be used if you return a response directly.

                    Read more about it in the
                    [FastAPI docs for Custom Response - HTML, Stream, File, others](https://fastapi.tiangolo.com/
                    advanced/custom-response/#redirectresponse).
                    """
                ),
            ] = Default(JSONResponse),
            name: Annotated[
                Optional[str],
                Doc(
                    """
                    Name for this *path operation*. Only used internally.
                    """
                ),
            ] = None,
            callbacks: Annotated[
                Optional[List[BaseRoute]],
                Doc(
                    """
                    List of *path operations* that will be used as OpenAPI callbacks.

                    This is only for OpenAPI documentation, the callbacks won't be used
                    directly.

                    It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                    Read more about it in the
                    [FastAPI docs for OpenAPI Callbacks](https://fastapi.tiangolo.com/advanced/openapi-callbacks/).
                    """
                ),
            ] = None,
            openapi_extra: Annotated[
                Optional[Dict[str, Any]],
                Doc(
                    """
                    Extra metadata to be included in the OpenAPI schema for this *path
                    operation*.

                    Read more about it in the
                    [FastAPI docs for Path Operation Advanced Configuration](https://fastapi.tiangolo.com/advanced/
                    path-operation-advanced-configuration/#custom-openapi-path-operation-schema).
                    """
                ),
            ] = None,
            generate_unique_id_function: Annotated[
                Callable[[APIRoute], str],
                Doc(
                    """
                    Customize the function used to generate unique IDs for the *path
                    operations* shown in the generated OpenAPI.

                    This is particularly useful when automatically generating clients or
                    SDKs for your API.

                    Read more about it in the
                    [FastAPI docs about how to Generate Clients](https://fastapi.tiangolo.com/advanced/generate-clients
                    /#custom-generate-unique-id-function).
                    """
                ),
            ] = Default(generate_unique_id),
            cache_policy: Annotated[
                Optional[CachePolicy],
                Doc(
                    """
                    TODO add description
                    """
                ),
            ] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """
        Add a *path operation* using an HTTP GET operation.

        ## Example

        ```python
        from fastapi import APIRouter, FastAPI

        app = FastAPI()
        router = APIRouter()

        @router.get('/items/')
        def read_items():
            return [{'name': 'Empanada'}, {'name': 'Arepa'}]

        app.include_router(router)
        ```
        """
        return self.api_route(
            path=path,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=['GET'],
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            callbacks=callbacks,
            openapi_extra=openapi_extra,
            generate_unique_id_function=generate_unique_id_function,
            cache_policy=cache_policy,
        )
